#notebook_panel { /* main background */
    background: #333;
    color: #f6f6f6;
}


div #notebook { /* centre the content */
    background: #FFFFF0;  /* IVORY */
    color: #333;
    width: 115ex;
    margin: auto;
    padding-left: 1em;
    padding-right: 1em;
    padding-top: 2ex;
}


div.text_cell_render {  /*Markdown cells */
    font-family: Computer Modern Roman;
}

/* remove big margins around title */
div.text_cell_render.border-box-sizing.rendered_html {
    margin: 0; /*What is this */
    padding: 0.05; /*Spacing between cells*/ 
    margin-left: 1ex;
}
div.cell.text_cell.border-box-sizing {
    margin: 0;
    padding: 0.05; 
}

.rendered_html h1 {
    margin: 0;
    padding: 0;
}
.rendered_html h2 {
    margin: 0;
    padding: 0;
}
.rendered_html h3 {
    margin: 0;
    padding: 0;
}

/* Width of all cells */
div.cell { /* set cell width to about 80 chars */
    width: 55em;

}

/* Set font of coding cells (DONT KNOW WHY BUT NEED THIS) */
.CodeMirror {
    /* Preferred font is bold sawasdee, but here are backups */
    font-family: "Sawasdee", "Lucida Sans Typewriter","Lucida Typewriter", 
         "Lucida Typewriter", "Courier New", Courier, monospace;;
    font-size: normal;
    font-weight: 700; 
}

div.cell.code_cell { /* Total code cell, in and out */
    background: #FFFFF0;
    border: none;
    padding-top: 1ex;
}

div.input_area { /* box around box with code ??? */
    border: none;
    background: #f5f5f5;
    border: 1px solid #fff;
    border-radius: 10px;
    padding-top: 0.5ex;
    padding-bottom: 0.5ex;
    padding-left: 0.5em;
}

div.input { /* box with code */
}


div.prompt { /* remove In/Out prompt */
    display: none;
}

div.output_subarea { /* remove margin in front of output */
    border: none;
}

div.cell.border-box-sizing.code_cell.running { 
    /* draw red dotted border around running cells */
    border: 3px dotted #f33;
}

/* header colours and fonts */
h1 { color: #444; }
h2 { color: #444; font-style: oblique}
h3 { color: #0000A0; font-weight: normal}
h4 { color: #444; font-weight: oblique; font-size: large}

